﻿using anbs_cp.Database.Interfaces;
using anbs_cp.Structs;

using Dapper;

using MySqlConnector;

namespace anbs_cp.Database.Classes;

public class MySqlEngine: IDbEngine
{
    /// <summary>
    /// Конструктор
    /// </summary>
    /// <param name="connectionString">Строка подключения базы данных</param>
    public MySqlEngine (string connectionString) => ConnectionString = connectionString;

    /// <summary>
    /// Строка подключения базы данных
    /// </summary>
    public string ConnectionString { get; set; }

    #region Небезопасные методы
    /// <summary>
    /// Выполняем команду
    /// </summary>
    /// <param name="sql">Запрос</param>
    /// <param name="model">Данные запроса</param>
    /// <returns>Успешно ли (есть ли хотя бы одна затронутая строки) выполнена команда</returns>
    public async Task<bool> ExecuteAsync (string sql, object model)
    {
        // Подключаемся к БД
        await using MySqlConnection connection = new(ConnectionString);

        // Открываем соединение
        await connection.OpenAsync();

        // Выполняем команду и выводим результат
        return await connection.ExecuteAsync(sql, model) > 0;
    }

    /// <summary>
    /// Запрос на получение одной строки
    /// </summary>
    /// <typeparam name="T">Тип передаваемого значения</typeparam>
    /// <param name="sql">Запрос</param>
    /// <param name="model">Данные запроса</param>
    /// <returns>Возвращает тип <see cref="T"/> или значение по умолчанию</returns>
    public async Task<T> QueryRowAsync<T> (string sql, object model)
    {
        // Подключаемся к БД
        await using MySqlConnection connection = new(ConnectionString);

        // Открываем соединение
        await connection.OpenAsync();

        // Выполняем запрос и выводим результат
        return await connection.QueryFirstOrDefaultAsync<T>(sql, model);
    }

    /// <summary>
    /// Запрос
    /// </summary>
    /// <typeparam name="T">Тип передаваемого значения</typeparam>
    /// <param name="sql">Запрос</param>
    /// <param name="model">Данные запроса</param>
    /// <returns>Возвращает массив типа <see cref="T"/> или значение по умолчанию</returns>
    public async Task<IEnumerable<T>> QueryAsync<T> (string sql, object model)
    {
        // Подключаемся к БД
        await using MySqlConnection connection = new(ConnectionString);

        // Открываем соединение
        await connection.OpenAsync();

        // Выполняем запрос и выводим результат
        return await connection.QueryAsync<T>(sql, model);
    }

    #endregion

    #region Получение

    /// <summary>
    /// Выборка данных из таблицы
    /// </summary>
    /// <typeparam name="T">Тип данных</typeparam>
    /// <param name="table">Имя таблицы</param>
    /// <param name="columns">Выбор столбцов таблицы, которые должны попасть в выборку (пустая -- все)</param>
    /// <param name="where">Массив значений столбцов, которые должны попасть в выборку</param>
    /// <param name="orderBy">Сортировка</param>
    /// <param name="limit">Сколько максимально элементов должно быть в выборке (0 -- без ограничения)</param>
    /// <returns>Массив типа <see cref="T"/></returns>
    public async Task<IEnumerable<T>> SelectAsync<T>(string table, string[] columns,
        List<KeyValue<string, string>> where, string? orderBy, int limit = 0)
    {
        // Формирую запрос
        string sql = $"""
                      SELECT
                      {(columns.Any() ? string.Join(", ", columns) : "*")}
                      FROM {table} 
                      """;

        // Создаю массив where параметров
        List<string> whereCondition = new();

        // Создаю массив значений
        List<string> paramsValues = new();

        // Если заданы параметры where
        if (where.Any())
        {
            // - то формирую массив where из where-запроса
            whereCondition.AddRange(where.Select(static whereItem => $"{whereItem.Key}=@{whereItem.Key}"));
            // - и значения параметров
            paramsValues.AddRange(where.Select(static whereItem => $"@{whereItem.Key}={whereItem.Value}"));
        }

        // Если массив where параметров не пуст
        if (whereCondition.Any())
            // - то добавляю к запросу
            sql += $" WHERE {string.Join(", ", whereCondition.ToArray())}";

        // Если добавлена сортировка
        if (orderBy != null)
            // - то добавляю сортировку к запросу
            sql += $" ORDER BY {orderBy}";

        // Если установлено запрашиваемое число элементов
        if (limit > 0)
            // - то вводим ограничение
            sql += $" LIMIT {limit}";

        // Осуществляю запрос и возвращаю результат
        return await QueryAsync<T>(sql, paramsValues.ToArray());
    }

    /// <summary>
    /// Выборка строки из таблицы
    /// </summary>
    /// <typeparam name="T">Тип данных</typeparam>
    /// <param name="table">Имя таблицы</param>
    /// <param name="columns">Выбор столбцов таблицы, которые должны попасть в выборку (пустая -- все)</param>
    /// <param name="where">Массив значений столбцов, которые должны попасть в выборку</param>
    /// <param name="orderBy">Сортировка</param>
    /// <returns>Тип <see cref="T"/> или null</returns>
    public async Task<T?> SelectRowAsync<T>(string table, string[] columns, List<KeyValue<string, string>> where,
        string? orderBy = null) => (await SelectAsync<T>(table, columns, where, orderBy, 1)).FirstOrDefault();

    #endregion

    #region Вычисление числа элементов в таблице
/// <summary>
/// Вычисляет число строк в таблице <paramref name="table"/>
/// </summary>
/// <param name="table">Таблица</param>
/// <returns>Число строк в таблице <paramref name="table"/></returns>
    public async Task<int> CountAsync(string table)
    {
        // Задаю запрос
        string sql = $"SELECT COUNT(*) FROM {table}";

        // Выполняю запрос
        return (await QueryAsync<int>(sql, new())).FirstOrDefault();
    }

    #endregion

    public async Task<TPit> InsertAsync<T, TPit> (string table, T entity, KeyValue<string, TPit> primaryIndex) => throw new NotImplementedException();

    public async Task<bool> InsertUpdateAsync<T> (string table, T entity) => throw new NotImplementedException();

    public async Task<bool> UpdateAsync<T> (string table, T entity, KeyValue<string, string> where) => throw new NotImplementedException();

    public async Task<bool> DeleteAsync<T> (string table, KeyValue<string, string> where) => throw new NotImplementedException();

    public async Task<bool> DeleteRowAsync<T> (string table, KeyValue<string, string> where) => throw new NotImplementedException();
}